/*
This sketch illustrates ideas from the course "2D Animation and Interaction"
available at www.imaginary-institute.com, and my book "Processing for
Artists and Designers," published by AK Peters. For more of my Processing
programs, visit www.glassner.com or www.imaginary-institute.com
---
Using Jared Tarbell's "SandPainter" technique, draw a series of concentric
rings. I darken them at the edges so they appear to be tubes. This looks
to me like the iris of an alien eye. 
---
I've re-programmed the SandPainter approach and wrapped it up in its own 
class, so it's easy to use in other projects. In fact, this program uses
a subclass of the SandPainter, which adds the effect of darkening the outer
edges of the zone being painted (in this case, a ring).
---
Andrew Glassner, version 2, April 2014
*/
// USER GLOBALS
int NumBands;                 // number of rings in the pattern
int PaintersPerBand;          // how many painters we use in each ring
TubePainter [] TubePainters;  // the global array of painters
color[] Palette;              // the list of colors to use for this drawing
void setup() {
  size(600, 400);  // a square canvas looks most eye-like, but any size is ok
  restart();       // create a new batch of painters
}
// clear the background and create a new random set of painters
void restart() {
  background(255);
  NumBands = int(random(3,7));            // how many rings are in the image
  PaintersPerBand = int(random(2,5));     // using multiple painters of different colors looks nice
  Palette = CanyonColors;                 // by default, use the canyon colors
  if (random(1.) > .5) Palette =  MetalColors;  // but half the time, use the metal colors
  int numTubePainters = NumBands * PaintersPerBand;   // how many painters we need
  TubePainters = new TubePainter[numTubePainters];    // allocate the memory
  for (int s=0; s<numTubePainters; s++) {
    TubePainters[s] = new TubePainter(randomColor(Palette)); // fill in each with starting values
  }
}
// Draw a random chunk of each ring. Find the inner and outer radii, and then
// pick a random pair of angles. Then draw the ring between those angles. Over
// time, the chunks overlap and accumulate in a pretty way to fill in the canvas.
void draw() {
  float cx = width/2;          // center x of the ring pattern
  float cy = height/2;         // center y
  float rmin = 0;              // radius of smallest ring
  float biggestSide = max(width, height);
  float rmax = sqrt(2)*biggestSide/2.; // radius of biggest ring
  float dr = (rmax-rmin)/NumBands;     // divide radius range into equal pieces
  for (int i=0; i<NumBands; i++) {     // for each band, run through all painters
    float r0 = rmin + (i * dr);        // get the inner radius of this band
    float r1 = r0 + dr;                // and the outer radius
    float t0 = random(0, TWO_PI);           // pick a starting angle for painting
    float t1 = t0 + random(PI/4, 7*PI/8);   // pick an ending angle for painting
    for (int b=0; b<PaintersPerBand; b++) { // march through all the painters for this band
      TubePainter s = TubePainters[b+(i*PaintersPerBand)]; // get the painter
      drawSandBand(cx, cy, r0, r1, t0, t1, s);             // and draw this section
    }
  }
}
// Find the two endpoints for the sand painter as we work our way around the ring.
// The ring is centered at (cx,cy), with radii r0 and r1, where r0 is the smaller of the two.
// We'll fill in the segment of the ring from angles t0 to t1 (as always, clockwise starting at 3 o'clock)
void drawSandBand(float cx, float cy, float r0, float r1, float t0, float t1, TubePainter s)
{
  float outlen = abs(t0-t1) * r1;  // 2pi r * (dtheta/2pi) - this is the length of the circle we're filling in
  int numPts = int(outlen);        // we'll draw that many short lines
  for (int i=0; i<numPts; i++) {
    float alfa = map(i, 0, numPts-1, 0, 1);  // alfa runs from 0 to 1 as we work our way through the segment
    float theta = lerp(t0, t1, alfa);        // lerp from t0 to t1 using alfa
    PVector p0 = new PVector(cx + (r0 * cos(theta)), cy + (r0 * sin(theta)));  // a little trig to find the two
    PVector p1 = new PVector(cx + (r1 * cos(theta)), cy + (r1 * sin(theta)));  //   points for this painter
    s.render(p0, p1);  // and draw some little piece of the line between those two points
  }
}
void mousePressed () {  // start over again
  restart();
}
//--------------- COLOR UTILITIES (usually in its own file)
// rusted desert metal. winter, new mexico. from Jared Tarbell of http://www.complexification.net
color[] MetalColors = {
  #3a242b, #3b2426, #352325, #836454, #7d5533, #8b7352, #b1a181, #a4632e, #bb6b33, #b47249,
  #ca7239, #d29057, #e0b87e, #d9b166, #f5eabe, #fcfadf, #d9d1b0, #fcfadf, #d1d1ca, #a7b1ac,
  #879a8c, #9186ad, #776a8e};
// perdenales canyon. fall, texas. from Jared Tarbell of http://www.complexification.net
color[] CanyonColors = {
  #f8f7f1, #6b6556, #a09c84, #908b7c, #79746e, #755d35, #937343, #9c6b4b, #ab8259, #aa8a61,
  #578375, #f0f6f2, #d0e0e5, #d7e5ec, #d3dfea, #c2d7e7, #a5c6e3, #a6cbe6, #adcbe5, #77839d,
  #d9d9b9, #a9a978, #727b5b, #6b7c4b, #546d3e, #47472e, #727b52, #898a6a, #919272, #AC623b,
  #cb6a33, #9d5c30, #843f2b, #652c2a, #7e372b, #403229, #47392b, #3d2626, #362c26, #57392c,
  #998a72, #864d36, #544732 };
color randomColor(color[] colors) { return colors[int(random(colors.length))]; }
// 80% of the time, return a nice color from the given list of colors.
// But 20% of the time, return either black or white with equal probability.
color randomBWColor(color[] colors) {
  if (random(1.) < .8) return randomColor(colors);  // pretty color
  if (random(1.) > .5) return(#000000);    // black
  return(#FFFFFF);                         // white
}
//--------------- SANDPAINTER CLASS (usually in its own file)
/*
This class is inspired by a technique due to Jared Tarbell at http://www.complexification.net
He callled it a "SandPainter," and I've kept the name. I've rewritten the algorithm, but it's
still Jared's method.
---
The essence is that we call render() to draw a line of dots that covers only a part of the range
between the two points it takes as arguments. We usually call render() repeatedly, moving the
endpoints a little bit (like a single pixel) between calls. The location and length of the line
of dots automatically moves around by small random amounts after each call. The result is that
we produce a kind of jittery band between the moving endpoints. Each little piece of the line is
transparent at the ends and opaque in the middle. Usually once we've gone through the entire range
of endpoints, we run through the same range again and again, letter the jittery bands overlap.
The result is an increasingly dense and attractive texture.
---
If the two endpoints each move by the same amount between calls, we'll get a thick band. But there
are no rules on how the endpoints move. One can stay fixed, producing a radial sweep, for example.
The results usually look best when successive calls to render() use endpoints that are each about
a pixel's distance from the previous call, but you can produce interesting effects by ignoring
that convention.
---
The usual approach has only two steps:
1. Create a SandPainter with a nice color and all the default values:
              SandPainter sp = new SandPainter(color(250, 200, 150));
2. Call render() over and over. Here's a big horizontal stroke from (100,100) to (300,200):
              while (true) { // accumulate new strokes on top of one another forever
                 for (int x=100; x<300; x++) {
                    PVector p0 = new PVector(x, 100); // top of band
                    PVector p1 = new PVector(x, 300); // bottom of band
                    sp.render(p0, p1);                // draw some dots in the band
                 }
              }
---
Your main controls:
setOpacity(float newOpacity):
              Set the overall opacity of the stroke, from 0 (transparent) to 1 (opaque). The
              default is 1. You can call this once when you first make the painter to control
              everything it draws. You can also call it more frequently, even once before each
              call to render(), allowing you to modulate the opacity of the jittery band along
              its length;
setHardEdges(boolean stayInLimits):
              This takes a boolean, which defaults to false. After each line of dots is drawn,
              the center and radius change a little bit in preparation for the next stroke. The
              center never goes beyond the endpoints passed to render(), but if it gets near
              one of them and the stroke's length is long enough, the dots that get drawn can
              appear outside of the range of the endpoints. This often looks nice, giving the
              edge of the stroke a nice fuzzy edge. If you want to force the dots to always
              stay within the endpoints given to render(), call this routine with a value of
              true(). You can change this setting as often as you want, so you can have some
              regions of your stroke with hard edges, and others that are soft.
---
Two technical controls: These next two calls let you set the numbers that control how the jiggly
line moves after each call to render(). They take numbers for arguments, and these values you need
to pick and tune by eye. They have nice defaults, so you may never need to call these. If you
do, I suggest starting with the default values (see the constructor for those), and then moving
them one at a time to see what they do and how the results look to you. The values should all
be between 0 and 1 for normal behavior, but nothing enforces that. These numbers are very
subjective; play with them until the results look good to you for your use.
---
setCenterInfo(centerWiggle):
              This takes a float that controls how much the center moves between calls to
              render(). It's a number that you pretty much have to pick by eye so that the
              stroke looks the way you like. It should be between 0 and 1, and generally ought
              to be pretty small so that the jiggly line has some continuity.
setRadiusInfo(float wiggleAmount, float minRadius, float maxRadius)
              This takes three floats. The first, wiggleAmount, controls how much the length
              of the line of dots changes between calls to render (the "radius" is half the length
              of the line, or the distance from the center to either endpoint). The value of
              minRadius sets the smallest length of the line of dots, and maxRadius sets the
              maximum length. These are both percentages of the distance between the endpoints,
              given in the range (0,1). Again, these are numbers you need to tune by eye to get
              just the look you're after.
---
One more technical control:
setDotDensity(float densityFactor)
              Because of the nature of the algorithm, the line of dots produced by render() will
              sometimes have a few single-pixel holes, and sometimes a few pixels will get written
              twice, making them a bit darker than their neighbors. This is rarely a problem, and
              to my eye it usually contributes to the overall texture in a very nice way. But you
              can control this by changing how many dots are produced along the line by calling
              setDotDensity() with a float specifying a new density. The default density is 1.0.
              A value greater than 1 will cause more dots to be drawn, which will eliminate the
              holes, but will produce more multiply-written pixels. A value less than 1 will cause
              fewer dots to be drawn, reducing the number of pixels that get written to twice, but
              increasing the number of holes. Note that the time it takes to render the stroke is
              proportional to the number of dots. So if you call setDotDensity(0.5) you'll get half
              the dots and require half the time, compared to the default of 1. A value of 2.0 will
              give you double the dots, but it will take twice as long as the default of 1.

I've broken out the steps for finding the color and opacity of each dot, and the update step
after each line of dots has been drawn, so that it's easy to make new variations on the algorithm.
Just subclass the SandPainter and replace (or extend) those routines in your own code. An
example subclass follows the main class.
---
Andrew Glassner, version 1 - April 2014
*/
class SandPainter {
  // user variables
  color clr;              // base color of the stroke
  float centerWiggle;     // how much the center wiggles on each step
  float radiusWiggle;     // how much the radius wiggles on each step
  float radiusMin, radiusMax;  // minimum and maximum radius values
  boolean hardEdges;      // true to make sure no dots go beyond the endpoints
  float opacity;          // overall opacity factor (0=transparent, 1=opaque)
  float dotDensity;       // factor to modify number of dots. >1 is denser but slower
  // state variables
  float center, radius;
  // create a new SandPainter with a given color and some nice default values
  SandPainter(color painterColor) {
    clr = painterColor;
    center = random(0.0, 1.0);
    radius = random(.05, .1);
    centerWiggle = .05;
    radiusWiggle = .025;
    radiusMin = .01;
    radiusMax = .2;
    hardEdges = false;
    dotDensity = 1.0;
    opacity = 1.0;
  }
  // set the amount by which the center wiggles on each step
  void setCenterInfo(float wiggleAmount) {
    centerWiggle = wiggleAmount;
  }
  // if true, then no dots will extend beyond the two points given to render()
  void setHardEdges(boolean stayInLimits) {
    hardEdges = stayInLimits;
  }
  // set the amount by which the radius wiggles on each step
  void setRadiusInfo(float wiggleAmount, float minRadius, float maxRadius) {
    radiusWiggle = wiggleAmount;
    radiusMin = minRadius;
    radiusMax = maxRadius;
  }
  // increase or decrease the number of dots to be drawn
  void setDotDensity(float densityFactor) {
    dotDensity = densityFactor;
  }
  // set the overall transparency of the stroke. 0=transparent, 1=opaque.
  void setOpacity(float newOpacity) {
    opacity = newOpacity;
  }
  // Draw a line of dots that cover some portion of the line from p0 to p1. The center of
  // the line of dots is given by "center", which runs from 0 to 1 between p0 and p1. The
  // distance from the center to either endpoint is "radius", also expressed as a percentage
  // of the distance from p0 to p1. The dots are transparent at both ends of the stroke but
  // smoothly become opaque at the center.
  void render(PVector p0, PVector p1) {
    PVector pA = p0.get();         // pA and pB are the endpoints of the line of dots
    pA.lerp(p1, center-radius);
    PVector pB = p0.get();
    pB.lerp(p1, center+radius);
    int numPts = int(dist(pA.x, pA.y, pB.x, pB.y));  // fill in most of the points between pA and pB
    numPts *= dotDensity;
    for (int i=0; i<numPts; i++) {                     // we'll draw one dot for each point
      float alfa = map(i, 0, numPts-1, 0, 1);          // alfa goes from 0 to 1 as we move from pA to pB
      float dotOpacity = getDotOpacity(center, alfa);  // get the opacity for this dot
      color dotColor = getDotColor(center, alfa);      // get the color for this dot
      color finalColor = color(red(dotColor), green(dotColor), blue(dotColor), opacity*dotOpacity*alpha(dotColor));
      // Known Processing bug: color(0,0,0,0) is drawn as opaque black, so don't draw it
      if (finalColor != 0) {
        stroke(finalColor);       // set the color to draw with
        PVector pnew = pA.get();  // locate the new dot by starting at pA,
        pnew.lerp(pB, alfa);      //    and moving alfa of the distance to pB
        point(pnew.x, pnew.y);
      }
    }
    update();
  }
  // The next three routines are broken out to make it easy to modify the algorithm for your own subclasses
  // center runs (0,1) between the overall stroke ends, alfa runs (0,1) over this smaller line of dots
  float getDotOpacity(float center, float alfa) {
    float opacity = map(sin(alfa*PI), 0, 1, 0, .25);  // 0 at ends, .25 in the center
    opacity = constrain(opacity, 0, .25);
    return(opacity);
  }
  // center runs (0,1) between the overall stroke ends, alfa runs (0,1) over this smaller line of dots
  color getDotColor(float center, float alfa) {
    return(clr);
  }
  // modify the values of center and radius for the next line of dots
  void update() {
    radius += radiusWiggle * random(-1, 1);            // wiggle the radius
    radius = constrain(radius, radiusMin, radiusMax);  // don't let the line get too short or too long
    center += centerWiggle * random(-1, 1);            // wiggle the center
    center = constrain(center, 0.0, 1.0);              // don't let center go outside the band
    if (hardEdges) center = constrain(center, radius, 1-radius);  // don't even let dots go outside the band
  }
}
//------------------- An example of how to subclass the SandPainter
// a TubePainter is darker at the top and bottom so it looks like a rounded tube
class TubePainter extends SandPainter {
  TubePainter(color painterClr) { // the required constructor for a subclass
    super(painterClr);
  }
  // go to black at top and bottom of the range
  color getDotColor(float center, float alfa) {
    float darken = map(sin(center * PI), 0, 1, .25, 1);
    darken = constrain(darken, 0, 1);  // just for safety
    color black = color(0);
    return lerpColor(black, clr, darken);
  }
}
